// Sample server for implementing the Chandy/Lamport snapshot algorithm.
// The server processes communicate with each other via messages (channels). A message is sent
// via the method send() and received via receive(). Both methods are inherited from
// BaseServer.

import java.rmi.*;
import java.util.Random;
import java.util.LinkedList;
import java.util.List;
import java.util.BitSet;


/**
 * Implementation of the server process.
 */
public class Server extends ServerBase
{
	private List<String> taskList;        // Number of tasks processed by this server
	private Random rnd;                   // Random number generator

	/**
	 * Initialize server process with number 'me', total number of processes is 'num'.
	 */
	public Server(int me, int num) throws RemoteException
	{
		super(me, num);
		rnd = new Random();
		taskList = new LinkedList<String>();
	}

	/**
	 * This method simulates the execution of a task.
	 * In this example, the server only sleeps for a while and then passes the task on to a
	 * randomly selected server with a 50% probability. This has the advantage that you can
	 * easily verify the result of the snapshot algorithm (the total number of tasks in the
	 * system is known and does not change).
	 */
	void doWork()
	{
		try {
			Thread.sleep(50 + rnd.nextInt(50));
		}
		catch (InterruptedException e) {
		}

        // If at least one task is available
		if (!taskList.isEmpty()) {
            // Select a task
 			String task = taskList.get(0);
            // Random number between 0 and 2*numServers-1
 			int n = rnd.nextInt(2*numServers);
			if (n < numServers) {
                // With a 50% probability, we get n < numServers. In this case, send a
                // task to the selected server and remove it from the 'taskList'.
				send(n, task);
				taskList.remove(0);
			}
		}
	}

    /**
     * Main loop of server process.
     */
 	void execute()
	{
		while(true) {
			// Message available?
			Message message = receive(false);
			if (message == null) {
				// No: do some work ...
				doWork();
			}
			else if (!message.getContents().equals("MARKER")) {
				// No marker message: add new task to 'taskList'
				taskList.add(message.getContents());
			}
			else {
				// Handling of MARKER messages
				// ... Add your solution here ... 
			}
		}
	}

	/**
	 * Main program: create the server object, register it with the RMI Registry and
	 * execute the main loop.
	 * Command line arguments: <Number of this server> <Total number of servers>
	 */
	public static void main(String args[])
	{
		try {
			int me = Integer.parseInt(args[0]);
			int num = Integer.parseInt(args[1]);
			Server obj = new Server(me, num);
			obj.execute();
		}
		catch (Exception e) {
			System.out.println("Error: " + e.getMessage());
			e.printStackTrace();
		}
	}
}
