// Base class for a message-based server.
// Functionality:
// - The server is an RMI object that registers in the constructor at the name service.
// - The method send(int to, String message) sends the message 'message' to the server
//   process with the number 'to'. If necessary, a lookup at the name service is
//   performed first. Then the message is entered into the queue of the server via the
//   remote method enqueueMessage(). 
// - A message is received using receive(). The operation returns null if no message is
//   received, otherwise a message object (consisting of message content and sender number).

import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.util.Queue;
import java.util.LinkedList;

/**
 * Implementation of the server process.
 */
public class ServerBase extends UnicastRemoteObject implements MessageReceiver
{
	private Queue<Message> messageQueue;  // A message queue for each channel
	private MessageReceiver[] server;     // Array with references to all server processes
	protected int serverNo;               // Number of this server process
	protected int numServers;             // Total number of server processes

	/**
	 * Initialize the server process with number 'me'; total number of processes is 'num'.
	 */
	public ServerBase(int me, int num) throws RemoteException
	{
		super();
		messageQueue = new LinkedList<Message>();
		server = new MessageReceiver[num];
		serverNo = me;
		numServers = num;

		try {
			Naming.bind("rmi://localhost/Server"+me, this);
			System.out.println("Server " + me + "/" + num + " registered");
		}
		catch (Exception e) {
			System.out.println("Error: " + e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Remote method for adding a message 'message' to the queue of a server process. The
	 * sending process has the number 'from' (-1 means client).
	 */
	public void enqueueMessage(String message, int from)
	{
		synchronized(messageQueue) {
			messageQueue.add(new Message(message, from));
			messageQueue.notify();
		}
	}

	/**
	 * Method for sending a message 'message' to the server process with number 'to'.
	 */
	public void send(int to, String message)
	{
		try {
			// If necessary, get the receiver's reference from the name service
			if (server[to] == null)
				server[to] = (MessageReceiver)Naming.lookup("rmi://localhost/Server" + to);
			// Enter the message into the receiver's message queue
			server[to].enqueueMessage(message, serverNo);
		}
		catch (Exception e) {
			System.out.println("Send error: " + e.getMessage());
		}
	}

	/**
	 * Method for querying and receiving a message. The 'block' parameter specifies whether to
	 * wait until a message arrives. If 'block == true' and there is no message, null is
	 * returned.
	 */
	public Message receive(boolean block)
	{
		Message message;
		synchronized(messageQueue) {
			while (block && messageQueue.isEmpty()) {
				try {
					messageQueue.wait();
				}
				catch (InterruptedException e) {
				}
			}
			message = messageQueue.poll();
		}
		return message;
	}
}
